/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include "dp.h"


/* since passing client_data didn't seem to work... */
DpStruct *draggedDpStruct;
static char *channelName;
static Widget warningDialog;
extern void getCompositeHeadTail(DpStruct *, DpStruct **, DpStruct **);

/* Function prototypes */
static void okCallback(Widget w, XtPointer client_data, XtPointer call_data);
static void dragDropFinish(Widget w, XtPointer client, XtPointer call);
static Boolean DragConvertProc(Widget w, Atom *selection, Atom *target,
  Atom *typeRtn, XtPointer *valueRtn, unsigned long *lengthRtn,
  int *formatRtn, unsigned long *max_lengthRtn, XtPointer client_data,
  XtRequestId *request_id);
void StartDrag(Widget w, XEvent *event);
void postWarningDialog(char *message);

static void dragDropFinish(Widget w, XtPointer client, XtPointer call)
{
  Widget sourceIcon;
  Pixmap pixmap;
  Arg args[2];

/* perform cleanup at conclusion of DND */
  XtSetArg(args[0],XmNsourcePixmapIcon,&sourceIcon);
  XtGetValues(w,args,1);

  XtSetArg(args[0],XmNpixmap,&pixmap);
  XtGetValues(sourceIcon,args,1);

  XFreePixmap(display,pixmap);
  XtDestroyWidget(sourceIcon);

}

static XtCallbackRec dragDropFinishCB[] = {
	{dragDropFinish,NULL},
	{NULL,NULL}
};


static Boolean DragConvertProc(Widget w, Atom *selection, Atom *target,
  Atom *typeRtn, XtPointer *valueRtn, unsigned long *lengthRtn,
  int *formatRtn, unsigned long *max_lengthRtn, XtPointer client_data,
  XtRequestId *request_id)
{
  XmString cString;
  char *cText, *passText;


  if (channelName != NULL) {
    if (*target != COMPOUND_TEXT) return(False);
    cString = XmStringCreateSimple(channelName);
    cText = XmCvtXmStringToCT(cString);
    passText = XtMalloc(strlen(cText)+1);
    memcpy(passText,cText,strlen(cText)+1);
   /* probably need this too */
    XmStringFree(cString);

   /* format the value for return */
    *typeRtn = COMPOUND_TEXT;
    *valueRtn = (XtPointer)passText;
    *lengthRtn = strlen(passText);
    *formatRtn = 8;	/* from example - related to #bits for data elements */
    return(True);
  } else {
    return(False);
  }

}




void StartDrag(Widget w, XEvent *event)
{
  Arg args[12];
  Cardinal n;
  Atom exportList[1];
  Widget sourceIcon;
  unsigned long fg, bg;
  XtPointer userData;
  XGCValues gcValues;
  unsigned long gcValueMask;
  Dimension width, height;
  Position x;
  DpStruct *ptr, *head, *tail;
  int numElements, j, k;
  XmFontContext context;
  XmFontList fontList;
  XmStringCharSet charset;
  static GC gc = NULL;
  static  XFontStruct *font = (XFontStruct *)NULL;
  Pixmap sourcePixmap = (Pixmap)NULL;

  draggedDpStruct = (DpStruct *)NULL;

  
/* since we are defining our own drag handling, we'll rely on
 *   dpStruct->channelL's userData field to store dpStruct from
 *   which we can get at the data we need to proceed
 */
  XtVaGetValues(w,XmNuserData,&userData,NULL);
  draggedDpStruct = (DpStruct *)userData;
  channelName = draggedDpStruct->channel;
  XtVaGetValues(draggedDpStruct->channelL,XmNheight,&height,
		XmNforeground,&fg,XmNbackground,&bg, NULL);

  if (draggedDpStruct->setPointTF != (Widget)NULL) {
    XtVaGetValues(draggedDpStruct->setPointTF,XmNx,&x,NULL);
    width = (Dimension)x;
  } else {
    width = (Dimension)(setPointX - deviceX);
  }

/* only do once since all channel labels have a consistent fontlist */
  if (font == (XFontStruct *)NULL) {
      XtVaGetValues(draggedDpStruct->channelL,XmNfontList,&fontList,NULL);
      XmFontListInitFontContext(&context,fontList);
      XmFontListGetNextFont(context,&charset,&font);
      XmFontListFreeFontContext(context);
  }

  if (draggedDpStruct != (DpStruct *)NULL) {
     if (!isCurrentlySelected(draggedDpStruct)) {
/* not already selected, therefore just move the selected element */
	numElements = 1;
	getCompositeHeadTail(draggedDpStruct,&head,&tail);
	ptr = head;
	while (ptr != tail && ptr != NULL) {
	   ptr = ptr->next;
	   numElements++;
	}
     } else {
/* picking an element that is already selected, therefore move all selected.. */
       numElements = 0;
       k = 0;
       while (currentlySelectedDpStructs[k] != (DpStruct *)NULL) {
	getCompositeHeadTail(currentlySelectedDpStructs[k],&head,&tail);
	ptr = head;
	while (ptr != tail && ptr != NULL) {
	   ptr = ptr->next;
	   numElements++;
	}
	numElements++;
	k++;
       }
     }

     sourcePixmap = XCreatePixmap(display,RootWindow(display,
	screenNum),(unsigned int)width,(unsigned int)(numElements*height),
	DefaultDepth(display,screenNum));
     if (gc == NULL) gc = XCreateGC(display,sourcePixmap,0,NULL);
     gcValueMask = GCForeground|GCBackground|GCFunction|GCFont;
     gcValues.foreground = bg;
     gcValues.background = bg;
     gcValues.function = GXcopy;
     gcValues.font = font->fid;
     XChangeGC(display,gc,gcValueMask,&gcValues);
     XFillRectangle(display,sourcePixmap,gc,0,0,width,numElements*height);
     XSetForeground(display,gc,fg);
     if (!isCurrentlySelected(draggedDpStruct)) {
/* render the single selected atomic/composite element */
	j = 0;
	getCompositeHeadTail(draggedDpStruct,&head,&tail);
	ptr = head;
	while (ptr != tail->next) {
	   if (ptr == draggedDpStruct && head != tail) {
	/* this is the actual selected element in a composite, render in
	   reverse video */
	      gcValueMask = GCForeground|GCBackground;
	      gcValues.foreground = bg;
	      gcValues.background = fg;
	      XChangeGC(display,gc,gcValueMask,&gcValues);
	      XDrawImageString(display,sourcePixmap,gc,
		(ptr->memberOfComposite ? (int)9*MARGIN : (int)MARGIN),
		(int)((j*height)+font->ascent),
		ptr->channel,strlen(ptr->channel));
	      gcValues.foreground = fg;
	      gcValues.background = bg;
	      XChangeGC(display,gc,gcValueMask,&gcValues);
	   } else {
	      XDrawString(display,sourcePixmap,gc,
		(ptr->memberOfComposite ? (int)9*MARGIN : (int)MARGIN),
		(int)((j*height)+font->ascent),
		ptr->channel,strlen(ptr->channel));
	   }
	   j++;
	   ptr = ptr->next;
	}
     } else {
/* render all currently selected elements */
       j = 0;
       k = 0;
       while (currentlySelectedDpStructs[k] != (DpStruct *)NULL) {
	getCompositeHeadTail(currentlySelectedDpStructs[k],&head,&tail);
	ptr = head;
	while (ptr != tail->next) {
	   if (ptr == draggedDpStruct) {
	/* this is the actual selected element, render in reverse video */
	      gcValueMask = GCForeground|GCBackground;
	      gcValues.foreground = bg;
	      gcValues.background = fg;
	      XChangeGC(display,gc,gcValueMask,&gcValues);
	      XDrawImageString(display,sourcePixmap,gc,
		(ptr->memberOfComposite ? (int)9*MARGIN : (int)MARGIN),
		(int)((j*height)+font->ascent),
		ptr->channel,strlen(ptr->channel));
	      gcValues.foreground = fg;
	      gcValues.background = bg;
	      XChangeGC(display,gc,gcValueMask,&gcValues);
	   } else {
	      XDrawString(display,sourcePixmap,gc,
		(ptr->memberOfComposite ? (int)9*MARGIN : (int)MARGIN),
		(int)((j*height)+font->ascent),
		ptr->channel,strlen(ptr->channel));
	   }
	   j++;
	   ptr = ptr->next;
	}
	k++;
       }
     }

  } else {

    channelName = NULL;
    return;

  }


  if (sourcePixmap != (Pixmap)NULL) {

/* use source widget as parent - can inherit visual attributes that way */
    n = 0;
    XtSetArg(args[n],XmNpixmap,sourcePixmap); n++;
    XtSetArg(args[n],XmNwidth,width); n++;
    XtSetArg(args[n],XmNheight,numElements*height); n++;
    XtSetArg(args[n],XmNdepth,DefaultDepth(display,screenNum)); n++;
    sourceIcon = XmCreateDragIcon(XtParent(w),"sourceIcon",args,n);

/* establish list of valid target types */
    exportList[0] = COMPOUND_TEXT;

    fg = BlackPixel(display,screenNum);
    bg = WhitePixel(display,screenNum);
    n = 0;
    XtSetArg(args[n],XmNexportTargets,exportList); n++;
    XtSetArg(args[n],XmNnumExportTargets,1); n++;
    XtSetArg(args[n],XmNdragOperations,XmDROP_COPY|XmDROP_MOVE); n++;
    XtSetArg(args[n],XmNconvertProc,DragConvertProc); n++;
    XtSetArg(args[n],XmNsourcePixmapIcon,sourceIcon); n++;
    XtSetArg(args[n],XmNcursorForeground,fg); n++;
    XtSetArg(args[n],XmNcursorBackground,bg); n++;
/*
    XtSetArg(args[n],XmNdragDropFinishCallback,dragDropFinishCB); n++;
*/
    XmDragStart(w,event,args,n);

  }
}



static void okCallback(Widget w, XtPointer client_data, XtPointer call_data)
{
    XtUnmanageChild(warningDialog);
}


void postWarningDialog(char *message)
{
  XmString xmString;
  Arg args[2];

  if (warningDialog == (Widget)NULL) {
    xmString = XmStringCreateLtoR(message,XmFONTLIST_DEFAULT_TAG);
    XtSetArg(args[0],XmNmessageString,xmString);
    XtSetArg(args[1],XmNtitle,"Warning");
    warningDialog = XmCreateWarningDialog(mainMW,
			"warningDialog", args, 2);
    XtUnmanageChild(XmMessageBoxGetChild(warningDialog,
                XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(warningDialog,
                XmDIALOG_CANCEL_BUTTON));
    XtAddCallback(warningDialog,XmNokCallback,okCallback,NULL);
  } else {
    xmString = XmStringCreateLtoR(message,XmFONTLIST_DEFAULT_TAG);
    XtSetArg(args[0],XmNmessageString,xmString);
    XtSetValues(warningDialog,args,1);
  }
  XtManageChild(warningDialog);
  XmStringFree(xmString);
}
