/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include "ctype.h"

#include "dp.h"
#include "cvtFast.h"

extern void freeCAresources(void);
extern void processSetPointConnectionEvent(struct connection_handler_args args);
extern void processReadbackConnectionEvent(struct connection_handler_args args);
extern void processStatusConnectionEvent(struct connection_handler_args args);

void exposeCallback(Widget  w, XtPointer client_data, XtPointer call_data);
void destroyBB1(Widget w, XtPointer client_data, XtPointer call_data);
void textFieldFloatVerifyCallback(Widget w, XtPointer client_data,  XtPointer call_data);
void setPointSend(Widget w, XtPointer client_data, XtPointer call_data);
void setPointLosingFocusCallback(Widget w, XtPointer client_data,  XtPointer call_data);
void createDisplayPage(char *filename);
void insertElementAfter(char *string, DpStruct *afterElement);
void renderStatus(DpStruct *dpStruct);


/*
 * dpStruct->DA's expose callback
 */
void exposeCallback(Widget  w, XtPointer client_data, XtPointer call_data)
{
    DpStruct *dpStruct= (DpStruct *)client_data;
    XmDrawingAreaCallbackStruct *cbs=(XmDrawingAreaCallbackStruct *)call_data;
    if (cbs->reason == XmCR_EXPOSE) {
	renderStatus(dpStruct);
    }
}


/*
 * mainBB1 destroy callback - just make sure selecting is taken care of
 */
void destroyBB1(Widget w, XtPointer client_data, XtPointer call_data)
{
  if (currentlySelectedDpStructs != (DpStruct **)NULL) {
    free((char *)currentlySelectedDpStructs);
    currentlySelectedDpStructs = (DpStruct **)NULL;
  }
  operRangeMaxColumns = NOMINAL_NCOLUMNS;
}


/*
 * float text field verify callback  (verify floating point input)
 */
void textFieldFloatVerifyCallback(Widget w, XtPointer client_data,  XtPointer call_data)
{
  int len;
  Boolean haveDecimalPoint;
  char *textString;
  XmTextVerifyCallbackStruct *cbs=(XmTextVerifyCallbackStruct *)call_data;

  len = 0;
  if (cbs->startPos < cbs->currInsert)	/* backspace*/
	return;

  len = XmTextFieldGetLastPosition(w);

  /* see if there is already a decimal point in the string */
  textString = XmTextFieldGetString(w);
  if (strchr(textString,(int)'.') != NULL)
	haveDecimalPoint = True;
  else
	haveDecimalPoint = False;
  XtFree(textString);

/* odd behavior for programmatic reset - if NULL event, then programmatic
	and bypass previous determinations regarding decimal point, etc */
  if (cbs->event == NULL) {
    len = 0;
    haveDecimalPoint = False;
  }

  /* if no (new) data e.g., setting to NULL string, simply return */
  if (cbs->text->ptr == NULL) return;

  /* check for leading sign (+/-) */
  if (len == 0) {	/* nothing there yet... therefore can add sign */
    if (!isdigit(cbs->text->ptr[0]) && cbs->text->ptr[0] != '+'
		&& cbs->text->ptr[0] != '-' && cbs->text->ptr[0] != '.') {
  /* not a digit or +/-/.,  move all chars down one and decrement length */
      int i;
      for (i = len; (i+1) < cbs->text->length; i++)
	cbs->text->ptr[i] = cbs->text->ptr[i+1];
      cbs->text->length--;
      len--;
    } else if (cbs->text->ptr[0] != '.') {
      haveDecimalPoint = True;
    }
  } else {
/* already a first character, therefore only digits or potential
 * decimal point allowed */

    for (len = 0; len < cbs->text->length; len++) {
    /* make sure all additions are digits (or the first decimal point) */
      if (!isdigit(cbs->text->ptr[len])) {
	 if (cbs->text->ptr[len] == '.' && !haveDecimalPoint) {
	  haveDecimalPoint = True;
	 } else {
	/* not a digit (or is another decimal point)  
	   - move all chars down one and decrement length */
	  int i;
	  for (i = len; (i+1) < cbs->text->length; i++)
	    cbs->text->ptr[i] = cbs->text->ptr[i+1];
	  cbs->text->length--;
	  len--;
	 }
      }
    }
  }
  if (cbs->text->length <= 0)
    cbs->doit = False;
}


/*
 * setPoint text field activate callback
 */
void setPointSend(Widget w, XtPointer client_data, XtPointer call_data)
{
  DpStruct *dpStruct;
  char *string;
  XmAnyCallbackStruct *cbs=(XmAnyCallbackStruct *)call_data;
    
  dpStruct = (DpStruct *)client_data;
  if (dpStruct != NULL) {
    string = XmTextFieldGetString(w);
    dpStruct->setPoint = atof(string);
/* if PV: reset statistics, do ca_put(), and mimic clearStatistics() */
    if (dpStruct->entryType == ProcessVariable) {
        dpStruct->count = 1;
        dpStruct->diffArrayIndex = -1;
        dpStruct->average = dpStruct->readback;
        dpStruct->averageDiffSquared = (double)0.0;
        dpStruct->maxDiff = (double)0.0;
        dpStruct->maxDiffAbs = (double)0.0;
    }
    XtFree(string);
/* otherwise, just do ca_put() and the rest will be handled by monitors */
    if (dpStruct->setPointChid != NULL) {
      if (ca_state(dpStruct->setPointChid) == cs_conn) {
        SEVCHK(ca_put(DBR_DOUBLE,dpStruct->setPointChid,&dpStruct->setPoint),
		"setPointSend: error in ca_put");
	ca_flush_io();
      } else {
	XmTextFieldSetString(w,"");
      }
    }
  }
}

/*
 * setPoint losing focus callback
 */
void setPointLosingFocusCallback(Widget w, XtPointer client_data,  XtPointer call_data)
{
  char string[40];
  DpStruct *dpStruct = (DpStruct *)client_data;
  XmTextVerifyCallbackStruct *cbs=( XmTextVerifyCallbackStruct *)call_data;

  if (dpStruct->setPointPreviouslyConnected) 
    cvtDoubleToCompactString(dpStruct->setPoint,string,
		dpStruct->setPointPrecision);
  else
    string[0] = '\0';
  XmTextFieldSetString(w,string);
}


void createDisplayPage(char *filename)
{
  FILE *stream;
  Arg args[8];
  DpStruct *dpStruct;
#define BUFFER_SIZE 256
#define STRING_SIZE 128
  char buffer[BUFFER_SIZE], charBuffer[STRING_SIZE];
  char *string;
  int numNames;
  Boolean notEOF;
  XmString labelString;
  int i, j, n, kk, status, numAtomics, numLines;
  Dimension width, nameWidth, labelWidth,
	rbWidth, rbHeight, spWidth, spHeight;
  Position rowHeight;
  Widget tempDA, savedSetPointTF=0, savedReadbackL=0;
  Boolean foundReadback, foundSetPoint, foundStatus;
  char *setPointPV=0, *readbackPV=0, *statusPV=0;

/*
 * bail out if cannot access the specified display page file
 */
  if (filename != NULL) {
    stream = fopen(filename,"r");
    if (stream == NULL) {
      fprintf(stderr,"\ncreateDisplayPage: can't read  %s",filename);
      return;
    }
  } else {
    return;
  }

  XDefineCursor(display,XtWindow(mainMW),watchCursor);

  if (listHead != (DpStruct *)NULL) freeCAresources();


/* initialize charBuffer (spaces for indentation of channel/pv names) */
  charBuffer[0] = ' '; charBuffer[1] = ' ';

/*
 * if already existing data, delete it and continue
 */
  XtUnmanageChild(mainMW);
  if (mainBB1 != NULL) {
    XtDestroyWidget(mainBB1);
  }


  /* create new BB (as child of mainMW's BB) */
  n = 0;
  XtSetArg(args[n],XmNshadowThickness,0); n++;
  XtSetArg(args[n],XmNresizePolicy,XmRESIZE_ANY); n++;
  mainBB1 = XmCreateBulletinBoard(mainBB,"mainBB1",args,n);
  XtAddCallback(mainBB1,XmNdestroyCallback,destroyBB1,NULL);


/*
 * now fill in data from file
 */

  numLines = 0;
  notEOF = True;
  rewind(stream);
  while(notEOF) {

    string = fgets(buffer,BUFFER_SIZE,stream);
    if (string == NULL) notEOF = False;
    if (notEOF) if (strlen(string) == 0) notEOF = False;

    if (notEOF) {

      strpbrk(string,"\n\0")[0] = '\0';

     /*
      * must be plain process variable...
      */
	/* append the new list element and update all relevant pointers */
	      dpStruct = (DpStruct *)malloc(sizeof(DpStruct));
	      dpStruct->next = NULL;
	      dpStruct->prev = listTail;
	      if (listHead == NULL) listHead = dpStruct;
	      if (listTail != NULL) listTail->next = dpStruct;
	      listTail = dpStruct;

	      dpStruct->entryType = ProcessVariable;
	      dpStruct->memberOfComposite = False;
	      strncpy(dpStruct->channel,string,CHANNEL_NAME_SIZE-1);
	      dpStruct->channel[CHANNEL_NAME_SIZE-1] = '\0';
 	      dpStruct->setPoint = 0.0;
	      dpStruct->readback = 0.0;
	      dpStruct->average = 0.0;
	      dpStruct->rmsDiff = 0.0;
	      dpStruct->maxDiff5 = 0.0;
	      dpStruct->maxDiff = 0.0;
	      dpStruct->maxDiffAbs = 0.0;
	      dpStruct->lopr = 0.0;
	      dpStruct->hopr = 0.0;
	      dpStruct->diffArrayIndex = -1;
	      dpStruct->averageDiffSquared = 0.0;
	      dpStruct->count = 0;
	      dpStruct->setPointPreviouslyConnected = False;
	      dpStruct->readbackPreviouslyConnected = False;
	      dpStruct->statusPreviouslyConnected = False;
	      dpStruct->modified = False;
	      dpStruct->setPointChid = NULL;
	      dpStruct->readbackChid = NULL;
	      dpStruct->statusChid = NULL;
	      dpStruct->setPointEvid = NULL;
	      dpStruct->readbackEvid = NULL;
	      dpStruct->statusEvid = NULL;
	      dpStruct->setPointString[0] = '\0';
	      dpStruct->readbackString[0] = '\0';
	      dpStruct->averageString[0] = '\0';
	      dpStruct->rmsDiffString[0] = '\0';
	      dpStruct->maxDiff5String[0] = '\0';
	      dpStruct->maxDiffString[0] = '\0';
	      dpStruct->unitsString[0] = '\0';
	      dpStruct->operRangeString[0] = '\0';
	      dpStruct->status = NoStatus;
	      dpStruct->foundSetPoint = False;
	      dpStruct->units[0] = '\0';
	      dpStruct->readbackChanged = True;
	      dpStruct->setPointChanged = True;
	      dpStruct->maxDiffChanged = True;
	  /* and create the widgets for this data */
	      n = 0;
	      XtSetArg(args[n],XmNuserData,(XtPointer)dpStruct); n++;
	      dpStruct->DA= XmCreateDrawingArea(mainBB1,"DA",args,n);
	      XtAddCallback(dpStruct->DA,XmNexposeCallback,
		exposeCallback,(XtPointer)dpStruct);
	      XtAddEventHandler(dpStruct->DA,ButtonPressMask,False,
		(XtEventHandler)selectRow,NULL);
	      tempDA = dpStruct->DA;
	      labelString= XmStringCreateSimple(dpStruct->channel);
	      n = 0;
	      XtSetArg(args[n],XmNlabelString,labelString); n++;
	      XtSetArg(args[n],XmNalignment,XmALIGNMENT_END); n++;
	      XtSetArg(args[n],XmNuserData,(XtPointer)dpStruct); n++;
	      dpStruct->channelL = XmCreateLabel(tempDA,"channelL",args,n);
	  /* add in drag/drop translations */
	      XtOverrideTranslations(dpStruct->channelL,parsedTranslations);

	      XmStringFree(labelString);

	      n = 0;
	      dpStruct->setPointTF= XmCreateTextField(tempDA,
			"setPointTF",args,n);
	      XmDropSiteUnregister(dpStruct->setPointTF);
	      XtAddCallback(dpStruct->setPointTF, XmNmodifyVerifyCallback,
		textFieldFloatVerifyCallback,(XtPointer)NULL);
	      XtAddCallback(dpStruct->setPointTF, XmNlosingFocusCallback,
		setPointLosingFocusCallback,
		(XtPointer)dpStruct);
	      XtAddCallback(dpStruct->setPointTF, XmNactivateCallback,
		setPointSend, (XtPointer)dpStruct);
	      savedSetPointTF = dpStruct->setPointTF;

	      n = 0;
	      dpStruct->readbackL= XmCreateLabel(tempDA,"simpleL",args,n);
	      savedReadbackL = dpStruct->readbackL;
	      dpStruct->averageL = XmCreateLabel(tempDA,"simpleL",args,n);
	      dpStruct->rmsDiffL = XmCreateLabel(tempDA,"simpleL",args,n);
	      dpStruct->maxDiff5L= XmCreateLabel(tempDA,"simpleL",args,n);
	      dpStruct->maxDiffL = XmCreateLabel(tempDA,"simpleL",args,n);
	      dpStruct->unitsL = XmCreateLabel(tempDA,"simpleL",args,n);
	      XtSetArg(args[n],XmNwidth,10*operRangeMaxColumns); n++;
	      dpStruct->operRangeL= XmCreateLabel(tempDA,
			"simpleL",args,n);

	      numLines++;
    }
  }

/* if didn't find anything simply return after remanaging the main window */
  if (listHead == (DpStruct *)NULL) {
     XUndefineCursor(display,XtWindow(mainMW));
     XtManageChild(mainMW);
     return;
  }



/*
 * create legend widgets and do the children management
 */

  /* entries in "rows" */
  nameWidth = 0;
  XtSetArg(args[0],XmNwidth,&width);
  dpStruct = listHead;
  while (dpStruct != (DpStruct *)NULL) {
    if (dpStruct->channelL != (Widget)NULL) {
	XtGetValues(dpStruct->channelL,args,1);
	nameWidth = MAX(width,nameWidth);
    }
    dpStruct = dpStruct->next;
  }

/* create device label in legend now */
  deviceX = 2*MARGIN;
  if (device == (Widget)NULL) {
    labelString = XmStringCreateSimple(legendString[DEVICE]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    device = XmCreateLabel(legendBB,"device",args,n);
    XtVaSetValues(device,XmNx,deviceX,XmNy,MARGIN,NULL);
    XtManageChild(device);
    XmStringFree(labelString);
  }
  XtVaGetValues(device,XmNwidth,&labelWidth,NULL);
  nameWidth = MAX(nameWidth,labelWidth);

/* update position of status LED */
  globalStatusX = nameWidth + 2*MARGIN;

  nameWidth += STATUS_DIAM + 5*MARGIN;

  /* get setPointTF width/height */
  XtSetArg(args[0],XmNwidth,&spWidth);
  XtSetArg(args[1],XmNheight,&spHeight);
  XtGetValues(savedSetPointTF,args,2);
/* make spWidth larger than setPointTF to allow more room for Rdbk fields */
  spWidth += 4*MARGIN;

  /* get readbackL width/height */
  XtSetArg(args[0],XmNwidth,&rbWidth);
  XtSetArg(args[1],XmNheight,&rbHeight);
  XtGetValues(savedReadbackL,args,2);
  rbWidth = (Dimension) MAX((int) spWidth+2*MARGIN,(int) rbWidth);
  rbWidth += MARGIN;
  rbWidth += 9;

/* move all the children of DA into their appropriate positions */
  dpStruct = listHead;
  while (dpStruct != (DpStruct *)NULL) {
   XtAddEventHandler(dpStruct->channelL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
   XtMoveWidget(dpStruct->channelL,MARGIN,MARGIN);
   XtManageChild(dpStruct->channelL);
   if (dpStruct->setPointTF != (Widget)NULL) {
       XtMoveWidget(dpStruct->setPointTF,nameWidth,MARGIN);
       XtManageChild(dpStruct->setPointTF);
   }
   if (dpStruct->readbackL != (Widget)NULL) {
       XtMoveWidget(dpStruct->readbackL,nameWidth+spWidth,MARGIN);
       XtAddEventHandler(dpStruct->readbackL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->readbackL);
   }
   if (dpStruct->averageL != (Widget)NULL) {
       XtMoveWidget(dpStruct->averageL,nameWidth+spWidth+rbWidth,MARGIN);
       XtAddEventHandler(dpStruct->averageL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->averageL);
   }
   if (dpStruct->rmsDiffL != (Widget)NULL) {
       XtMoveWidget(dpStruct->rmsDiffL,nameWidth+spWidth+2*rbWidth,MARGIN);
       XtAddEventHandler(dpStruct->rmsDiffL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtVaSetValues(dpStruct->rmsDiffL,XmNx,nameWidth+spWidth+2*rbWidth,NULL);
       XtManageChild(dpStruct->rmsDiffL);
   }
   if (dpStruct->maxDiff5L != (Widget)NULL) {
       XtMoveWidget(dpStruct->maxDiff5L,nameWidth+spWidth+3*rbWidth,MARGIN);
       XtAddEventHandler(dpStruct->maxDiff5L,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->maxDiff5L);
   }
   if (dpStruct->maxDiffL != (Widget)NULL) {
       XtMoveWidget(dpStruct->maxDiffL,nameWidth+spWidth+4*rbWidth,MARGIN);
       XtAddEventHandler(dpStruct->maxDiffL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->maxDiffL);
   }
   if (dpStruct->unitsL != (Widget)NULL) {
       XtMoveWidget(dpStruct->unitsL,nameWidth+spWidth+5*rbWidth,MARGIN);
       XtAddEventHandler(dpStruct->unitsL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->unitsL);
   }
   if (dpStruct->operRangeL != (Widget)NULL) {
       XtMoveWidget(dpStruct->operRangeL,nameWidth+spWidth+6*rbWidth,MARGIN);
       XtAddEventHandler(dpStruct->operRangeL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->operRangeL);
   }
   dpStruct = dpStruct->next;

  }

  /* size, position and manage "rows" (DAs) */
  rowHeight = MAX(spHeight,rbHeight);
  XtSetArg(args[0],XmNheight,(Dimension)rowHeight);
  j = 1;
  dpStruct = listHead;
  while (dpStruct != NULL) {
    XtSetArg(args[1],XmNy,(Position)((j-1)*(rowHeight+MOVE_MARGIN)));
    XtSetValues(dpStruct->DA,args,2);
    XtManageChild(dpStruct->DA);
    dpStruct = dpStruct->next;
    j++;
  }

  XtManageChild(mainBB1);
  XtVaGetValues(mainBB1,XmNwidth,&width,NULL);

/* make sure all DAs are full size even if not many children */
  XtSetArg(args[0],XmNwidth,width);
  dpStruct = listHead;
  while (dpStruct != NULL) {
    XtSetValues(dpStruct->DA,args,1);
    dpStruct = dpStruct->next;
  }
 
  XtVaSetValues(mainShell,XmNwidth,width+MW_MARGIN,NULL);


  if (setPoint != (Widget)NULL) {
     setPointX = nameWidth;
     rdbkX = nameWidth+spWidth;
     averageX = nameWidth+spWidth+rbWidth;
     rmsX = nameWidth+spWidth+2*rbWidth;
     diff5X = nameWidth+spWidth+3*rbWidth;
     diffX = nameWidth+spWidth+4*rbWidth;
     unitsX = nameWidth+spWidth+5*rbWidth;
     operRangeX = nameWidth+spWidth+6*rbWidth;
     XtMoveWidget(device,deviceX,MARGIN);
     XtMoveWidget(setPoint,setPointX,MARGIN);
     XtMoveWidget(rdbk,rdbkX,MARGIN);
     XtMoveWidget(average,averageX,MARGIN);
     XtMoveWidget(rms,rmsX,MARGIN);
     XtMoveWidget(diff5,diff5X,MARGIN);
     XtMoveWidget(diff,diffX,MARGIN);
     XtMoveWidget(units,unitsX,MARGIN);
     XtMoveWidget(operRange,operRangeX,MARGIN);
  } else {
    setPointX = nameWidth;
    rdbkX = nameWidth+spWidth;
    averageX = nameWidth+spWidth+rbWidth;
    rmsX = nameWidth+spWidth+2*rbWidth;
    diff5X = nameWidth+spWidth+3*rbWidth;
    diffX = nameWidth+spWidth+4*rbWidth;
    unitsX = nameWidth+spWidth+5*rbWidth;
    operRangeX = nameWidth+spWidth+6*rbWidth;

/* (place all labels over their corresponding rows) */

/* setPoint */
    labelString = XmStringCreateSimple(legendString[SET_POINT]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,setPointX); n++;
    setPoint = XmCreateLabel(legendBB,"setPoint",args,n);
    XtVaGetValues(setPoint,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(setPoint,XmNx,setPointX,XmNy,MARGIN,NULL);
    XtManageChild(setPoint);
    XmStringFree(labelString);

/* rdbk */
    labelString = XmStringCreateSimple(legendString[RDBK]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,rdbkX); n++;
    rdbk = XmCreateLabel(legendBB,"rdbk",args,n);
    XtVaGetValues(rdbk,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(rdbk,XmNx,rdbkX,XmNy,MARGIN,NULL);
    XtManageChild(rdbk);
    XmStringFree(labelString);

/* average */
    labelString = XmStringCreateSimple(legendString[AVERAGE]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,averageX); n++;
    average = XmCreateLabel(legendBB,"average",args,n);
    XtVaGetValues(average,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(average,XmNx,averageX,XmNy,MARGIN,NULL);
    XtManageChild(average);
    XmStringFree(labelString);

/* rms */
    labelString = XmStringCreateSimple(legendString[RMS]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,rmsX); n++;
    rms = XmCreateLabel(legendBB,"rms",args,n);
    XtVaGetValues(rms,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(rms,XmNx,rmsX,XmNy,MARGIN,NULL);
    XtManageChild(rms);
    XmStringFree(labelString);

/* diff5 */
    labelString = XmStringCreateSimple(legendString[DIFF5]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,diff5X); n++;
    diff5 = XmCreateLabel(legendBB,"diff5",args,n);
    XtVaGetValues(diff5,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(diff5,XmNx,diff5X,XmNy,MARGIN,NULL);
    XtManageChild(diff5);
    XmStringFree(labelString);

/* diff */
    labelString = XmStringCreateSimple(legendString[DIFF]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,diffX); n++;
    diff = XmCreateLabel(legendBB,"diff",args,n);
    XtVaGetValues(diff,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(diff,XmNx,diffX,XmNy,MARGIN,NULL);
    XtManageChild(diff);
    XmStringFree(labelString);

/* units */
    labelString = XmStringCreateSimple(legendString[UNITS]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,unitsX); n++;
    units = XmCreateLabel(legendBB,"units",args,n);
    XtVaGetValues(units,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(units,XmNx,unitsX,XmNy,MARGIN,NULL);
    XtManageChild(units);
    XmStringFree(labelString);

/* operating range */
    labelString = XmStringCreateSimple(legendString[OPER_RANGE]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,operRangeX); n++;
    operRange = XmCreateLabel(legendBB,"operRange",args,n);
    XtVaGetValues(operRange,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(operRange,XmNx,operRangeX,XmNy,MARGIN,NULL);
    XtManageChild(operRange);
    XmStringFree(labelString);

  }

  XtManageChild(mainMW);
  XFlush(display);


/* add CA connection handling */
  dpStruct = listHead;

  while (dpStruct != (DpStruct *)NULL) {

  if (dpStruct->entryType == ProcessVariable) {
/*
 * ProcessVariable
 */
      dpStruct->foundSetPoint = False;
      /* not really treating setPoint as monitorable, but need info */
      SEVCHK(ca_build_and_connect(dpStruct->channel,
		TYPENOTCONN,0,
		&(dpStruct->setPointChid),NULL,processSetPointConnectionEvent,
		NULL), "createDisplayPage: error in ca_build_and_connect");
      if (dpStruct->setPointChid != NULL)
		ca_set_puser(dpStruct->setPointChid,dpStruct);
      SEVCHK(ca_build_and_connect(dpStruct->channel,
		TYPENOTCONN,0,
		&(dpStruct->readbackChid),NULL,processReadbackConnectionEvent,
		NULL), "createDisplayPage: error in ca_build_and_connect");
      if (dpStruct->readbackChid != NULL)
		ca_set_puser(dpStruct->readbackChid,dpStruct);
    }

    dpStruct = dpStruct->next;

   }

   ca_pend_event(10.0*CA_PEND_EVENT_TIME);

/* and scroll the window */
   if (listHead != (DpStruct *)NULL)
	XmScrollVisible(mainMW,listHead->channelL,2,2);

   XUndefineCursor(display,XtWindow(mainMW));

}



/*
 * function to append an element (row) to the display
 *   arguments:  char *string = name to be resolved and added
 */
void insertElementAfter(char *string, DpStruct *afterElement)
{
  Arg args[8];
  DpStruct *dpStruct,  *localListHead, *localListTail,
	 *beforeElement;
#define BUFFER_SIZE 256
#define STRING_SIZE 128
  char charBuffer[STRING_SIZE];
  int numNames;
  Widget scrollToWidget;
  XmString labelString;
  int i, j, n, kk, status, numAtomics, numLines;
  Dimension width, nameWidth, labelWidth, 
	rbWidth, rbHeight, spWidth, spHeight;
  Position rowHeight, afterY;
  Widget tempDA, savedSetPointTF=0, savedReadbackL=0;
  Boolean foundReadback, foundSetPoint, foundStatus;
  char *setPointPV=0, *readbackPV=0, *statusPV=0;


  if (string == NULL) return;
  if (strlen(string) <= (size_t)0) return;

  scrollToWidget = (Widget)NULL;

  XDefineCursor(display,XtWindow(mainMW),watchCursor);

/* initialize charBuffer (spaces for indentation of channel/pv names) */
  charBuffer[0] = ' '; charBuffer[1] = ' ';


/* element after which to append elements (afterElement), and element
 *  before which (beforeElement) to insert elements */
  if (afterElement != NULL) beforeElement = afterElement->next;
  else beforeElement = NULL;
  localListHead = NULL;
  localListTail = NULL;

  if (mainBB1 == (Widget)NULL) {
  /* create new BB (as child of mainMW's BB) */
    n = 0;
    XtSetArg(args[n],XmNshadowThickness,0); n++;
    XtSetArg(args[n],XmNresizePolicy,XmRESIZE_ANY); n++;
    mainBB1 = XmCreateBulletinBoard(mainBB,"mainBB1",args,n);
    XtAddCallback(mainBB1,XmNdestroyCallback,destroyBB1,NULL);
  }

/*
 * now fill in data
 */

  numLines = 0;

     /*
      * must be plain process variable...
      */
/* append the new localList element and update all relevant pointers */
      dpStruct = (DpStruct *)calloc(1,sizeof(DpStruct));
      dpStruct->next = NULL;
      dpStruct->prev = localListTail;
      if (localListHead == NULL) localListHead = dpStruct;
      if (localListTail != NULL) localListTail->next = dpStruct;
      localListTail = dpStruct;

      dpStruct->entryType = ProcessVariable;
      dpStruct->memberOfComposite = False;
      strncpy(dpStruct->channel,string,CHANNEL_NAME_SIZE-1);
      dpStruct->channel[CHANNEL_NAME_SIZE-1] = '\0';
 	      dpStruct->setPoint = 0.0;
      dpStruct->readback = 0.0;
      dpStruct->average = 0.0;
      dpStruct->rmsDiff = 0.0;
      dpStruct->maxDiff5 = 0.0;
      dpStruct->maxDiff = 0.0;
      dpStruct->maxDiffAbs = 0.0;
      dpStruct->lopr = 0.0;
      dpStruct->hopr = 0.0;
      dpStruct->diffArrayIndex = -1;
      dpStruct->averageDiffSquared = 0.0;
      dpStruct->count = 0;
      dpStruct->setPointPreviouslyConnected = False;
      dpStruct->readbackPreviouslyConnected = False;
      dpStruct->statusPreviouslyConnected = False;
      dpStruct->modified = False;
      dpStruct->setPointChid = NULL;
      dpStruct->readbackChid = NULL;
      dpStruct->statusChid = NULL;
      dpStruct->setPointEvid = NULL;
      dpStruct->readbackEvid = NULL;
      dpStruct->statusEvid = NULL;
      dpStruct->status = NoStatus;
      dpStruct->foundSetPoint = False;
      dpStruct->units[0] = '\0';
      dpStruct->readbackChanged = True;
      dpStruct->setPointChanged = True;
      dpStruct->maxDiffChanged = True;
  /* and create the widgets for this data */
      n = 0;
      XtSetArg(args[n],XmNresizePolicy,XmRESIZE_GROW); n++;
      XtSetArg(args[n],XmNuserData,(XtPointer)dpStruct); n++;
      dpStruct->DA= XmCreateDrawingArea(mainBB1,"DA",args,n);
      XtAddCallback(dpStruct->DA,XmNexposeCallback,
      exposeCallback,(XtPointer)dpStruct);
      XtAddEventHandler(dpStruct->DA,ButtonPressMask,False,
      (XtEventHandler)selectRow,NULL);
      tempDA = dpStruct->DA;
      labelString= XmStringCreateSimple(dpStruct->channel);
      n = 0;
      XtSetArg(args[n],XmNlabelString,labelString); n++;
      XtSetArg(args[n],XmNalignment,XmALIGNMENT_END); n++;
      XtSetArg(args[n],XmNuserData,(XtPointer)dpStruct); n++;
      dpStruct->channelL = XmCreateLabel(tempDA,"channelL",args,n);
      if (scrollToWidget == (Widget)NULL)
      	scrollToWidget = dpStruct->channelL;
  /*add in drag/drop translations */
      XtOverrideTranslations(dpStruct->channelL,parsedTranslations);
      XmStringFree(labelString);

      n = 0;
      dpStruct->setPointTF= XmCreateTextField(tempDA,
      	"setPointTF",args,n);
      XmDropSiteUnregister(dpStruct->setPointTF);
      XtAddCallback(dpStruct->setPointTF, XmNmodifyVerifyCallback,
      textFieldFloatVerifyCallback,(XtPointer)NULL);
      XtAddCallback(dpStruct->setPointTF, XmNlosingFocusCallback,
      setPointLosingFocusCallback,
      (XtPointer)dpStruct);
      XtAddCallback(dpStruct->setPointTF, XmNactivateCallback,
      setPointSend, (XtPointer)dpStruct);
      savedSetPointTF = dpStruct->setPointTF;

      n = 0;
      dpStruct->readbackL= XmCreateLabel(tempDA,"simpleL",args,n);
      savedReadbackL = dpStruct->readbackL;
      dpStruct->averageL = XmCreateLabel(tempDA,"simpleL",args,n);
      dpStruct->rmsDiffL = XmCreateLabel(tempDA,"simpleL",args,n);
      dpStruct->maxDiff5L= XmCreateLabel(tempDA,"simpleL",args,n);
      dpStruct->maxDiffL = XmCreateLabel(tempDA,"simpleL",args,n);
      dpStruct->unitsL = XmCreateLabel(tempDA,"simpleL",args,n);
      XtSetArg(args[n],XmNwidth,10*operRangeMaxColumns); n++;
      dpStruct->operRangeL= XmCreateLabel(tempDA,"simpleL",args,n);

      numLines++;



  /* merge localList into main list */
  if (afterElement != NULL) afterElement->next = localListHead;
  else listHead = localListHead;
  localListHead->prev = afterElement;
  if (beforeElement != NULL) beforeElement->prev = localListTail;
  else listTail = localListTail;
  localListTail->next = beforeElement;



/*
 * create legend widgets and do the children management
 */

  /* entries in "rows" */
  nameWidth = 0;
  XtSetArg(args[0],XmNwidth,&width);
  dpStruct = listHead;
  while (dpStruct != NULL) {
    if (dpStruct->channelL != (Widget)NULL) {
	XtGetValues(dpStruct->channelL,args,1);
	nameWidth = MAX(width,nameWidth);
    }
    dpStruct = dpStruct->next;
  }

/* create device label in legend now */
  deviceX = 2*MARGIN;
  if (device == (Widget)NULL) {
    labelString = XmStringCreateSimple(legendString[DEVICE]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    device = XmCreateLabel(legendBB,"device",args,n);
    XtVaSetValues(device,XmNx,deviceX,XmNy,MARGIN,NULL);
    XtManageChild(device);
    XmStringFree(labelString);
  }
  XtVaGetValues(device,XmNwidth,&labelWidth,NULL);
  nameWidth = MAX(nameWidth,labelWidth);

/* update position of status LED */
  globalStatusX = nameWidth + 2*MARGIN;

  nameWidth += STATUS_DIAM + 5*MARGIN;

  /* get setPointTF width/height */
  XtSetArg(args[0],XmNwidth,&spWidth);
  XtSetArg(args[1],XmNheight,&spHeight);
  XtGetValues(savedSetPointTF,args,2);
/* make spWidth larger than setPointTF to allow more room for Rdbk fields */
  spWidth += 4*MARGIN;

  /* get readbackL width/height */
  XtSetArg(args[0],XmNwidth,&rbWidth);
  XtSetArg(args[1],XmNheight,&rbHeight);
  XtGetValues(savedReadbackL,args,2);
  rbWidth = (Dimension) MAX((int) spWidth+2*MARGIN,(int) rbWidth);
  rbWidth += MARGIN;
  rbWidth += 9;

/* move all the children of DA into their appropriate positions */
  dpStruct = localListHead;
  while (dpStruct != localListTail->next) {
   XtAddEventHandler(dpStruct->channelL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
   XtMoveWidget(dpStruct->channelL,MARGIN,MARGIN);
   XtManageChild(dpStruct->channelL);
   if (dpStruct->setPointTF != (Widget)NULL) {
       XtMoveWidget(dpStruct->setPointTF,nameWidth,MARGIN);
       XtManageChild(dpStruct->setPointTF);
   }
   if (dpStruct->readbackL != (Widget)NULL) {
       XtMoveWidget(dpStruct->readbackL,nameWidth+spWidth,MARGIN);
       XtAddEventHandler(dpStruct->readbackL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->readbackL);
   }
   if (dpStruct->averageL != (Widget)NULL) {
       XtMoveWidget(dpStruct->averageL,nameWidth+spWidth+rbWidth,MARGIN);
       XtAddEventHandler(dpStruct->averageL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->averageL);
   }
   if (dpStruct->rmsDiffL != (Widget)NULL) {
       XtMoveWidget(dpStruct->rmsDiffL,nameWidth+spWidth+2*rbWidth,MARGIN);
       XtAddEventHandler(dpStruct->rmsDiffL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->rmsDiffL);
   }
   if (dpStruct->maxDiff5L != (Widget)NULL) {
       XtMoveWidget(dpStruct->maxDiff5L,nameWidth+spWidth+3*rbWidth,MARGIN);
       XtAddEventHandler(dpStruct->maxDiff5L,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->maxDiff5L);
   }
   if (dpStruct->maxDiffL != (Widget)NULL) {
       XtMoveWidget(dpStruct->maxDiffL,nameWidth+spWidth+4*rbWidth,MARGIN);
       XtAddEventHandler(dpStruct->maxDiffL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->maxDiffL);
   }
   if (dpStruct->unitsL != (Widget)NULL) {
       XtMoveWidget(dpStruct->unitsL,nameWidth+spWidth+5*rbWidth,MARGIN);
       XtAddEventHandler(dpStruct->unitsL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->unitsL);
   }
   if (dpStruct->operRangeL != (Widget)NULL) {
       XtMoveWidget(dpStruct->operRangeL,nameWidth+spWidth+6*rbWidth,MARGIN);
       XtAddEventHandler(dpStruct->operRangeL,ButtonPressMask,False,
	(XtEventHandler)selectRow,NULL);
       XtManageChild(dpStruct->operRangeL);
   }
   dpStruct = dpStruct->next;

  }

  /* size, position and manage "rows" (DAs) */
  rowHeight = MAX(spHeight,rbHeight);
  XtSetArg(args[0],XmNheight,(Dimension)rowHeight);
  if (afterElement != NULL) {
    XtVaGetValues(afterElement->DA,XmNy,&afterY,NULL);
    afterY += rowHeight+MOVE_MARGIN;		/* like below */
  } else afterY = 0;

  j = 1;
  dpStruct = localListHead;
  while (dpStruct != NULL) {	/* want to move ALL of the children after...*/
    XtSetArg(args[1],XmNy,(Position)(afterY + (j-1)*(rowHeight+MOVE_MARGIN)));
    XtSetValues(dpStruct->DA,args,2);
    XtManageChild(dpStruct->DA);
    dpStruct = dpStruct->next;
    j++;
  }


/* scoot everything over if channel name sizes changed (don't use XtMoveWidget()
 *   because resize sensitive parents like DA really do care... */
  if (nameWidth != (Dimension)setPointX) {
   dpStruct = listHead;
   while (dpStruct != NULL) {
     if (dpStruct->setPointTF != (Widget)NULL)
       XtVaSetValues(dpStruct->setPointTF,XmNx,nameWidth,NULL);
     if (dpStruct->readbackL != (Widget)NULL)
       XtVaSetValues(dpStruct->readbackL,XmNx,nameWidth+spWidth,NULL);
     if (dpStruct->averageL != (Widget)NULL)
       XtVaSetValues(dpStruct->averageL,XmNx,nameWidth+spWidth+rbWidth,NULL);
     if (dpStruct->rmsDiffL != (Widget)NULL)
       XtVaSetValues(dpStruct->rmsDiffL,XmNx,nameWidth+spWidth+2*rbWidth,NULL);
     if (dpStruct->maxDiff5L != (Widget)NULL)
       XtVaSetValues(dpStruct->maxDiff5L,XmNx,nameWidth+spWidth+3*rbWidth,
		NULL);
     if (dpStruct->maxDiffL != (Widget)NULL)
       XtVaSetValues(dpStruct->maxDiffL,XmNx,nameWidth+spWidth+4*rbWidth,NULL);
     if (dpStruct->unitsL != (Widget)NULL)
       XtVaSetValues(dpStruct->unitsL,XmNx,nameWidth+spWidth+5*rbWidth,NULL);
     if (dpStruct->operRangeL != (Widget)NULL)
       XtVaSetValues(dpStruct->operRangeL,XmNx,nameWidth+spWidth+6*rbWidth,
		NULL);

     dpStruct = dpStruct->next;
   }
  /* always scoot labels over if nameWidth != setPointX */
   deviceX = 2*MARGIN;
   setPointX = nameWidth;
   rdbkX = nameWidth+spWidth;
   averageX = nameWidth+spWidth+rbWidth;
   rmsX = nameWidth+spWidth+2*rbWidth;
   diff5X = nameWidth+spWidth+3*rbWidth;
   diffX = nameWidth+spWidth+4*rbWidth;
   unitsX = nameWidth+spWidth+5*rbWidth;
   operRangeX = nameWidth+spWidth+6*rbWidth;
   if (device != (Widget)NULL) XtMoveWidget(device,deviceX,MARGIN);
   if (setPoint != (Widget)NULL) XtMoveWidget(setPoint,setPointX,MARGIN);
   if (rdbk != (Widget)NULL) XtMoveWidget(rdbk,rdbkX,MARGIN);
   if (average != (Widget)NULL) XtMoveWidget(average,averageX,MARGIN);
   if (rms != (Widget)NULL) XtMoveWidget(rms,rmsX,MARGIN);
   if (diff5 != (Widget)NULL) XtMoveWidget(diff5,diff5X,MARGIN);
   if (diff != (Widget)NULL) XtMoveWidget(diff,diffX,MARGIN);
   if (units != (Widget)NULL) XtMoveWidget(units,unitsX,MARGIN);
   if (operRange != (Widget)NULL) XtMoveWidget(operRange,operRangeX,MARGIN);

/* force redraw for status, etc */
   dpStruct = listHead;
   while (dpStruct != (DpStruct *)NULL) {
     if (dpStruct->DA != (Widget)NULL) {
     /* probably could bail out as soon as one fully invisible is found... */
	if (XmGetVisibility(dpStruct->DA) != XmVISIBILITY_FULLY_OBSCURED)
	    XClearWindow(display,XtWindow(dpStruct->DA));
     }
     dpStruct = dpStruct->next;
   }

  } /* end if moved... */


/*
 *  now fill in legend if not already done so
 */
  if (setPoint == (Widget)NULL) {

    deviceX = 2*MARGIN;
    setPointX = nameWidth;
    rdbkX = nameWidth+spWidth;
    averageX = nameWidth+spWidth+rbWidth;
    rmsX = nameWidth+spWidth+2*rbWidth;
    diff5X = nameWidth+spWidth+3*rbWidth;
    diffX = nameWidth+spWidth+4*rbWidth;
    unitsX = nameWidth+spWidth+5*rbWidth;
    operRangeX = nameWidth+spWidth+6*rbWidth;

/* (place all labels over their corresponding rows) */

/* setPoint */
    labelString = XmStringCreateSimple(legendString[SET_POINT]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,setPointX); n++;
    setPoint = XmCreateLabel(legendBB,"setPoint",args,n);
    XtVaGetValues(setPoint,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(setPoint,XmNx,setPointX,XmNy,MARGIN,NULL);
    XtManageChild(setPoint);
    XmStringFree(labelString);

/* rdbk */
    labelString = XmStringCreateSimple(legendString[RDBK]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,rdbkX); n++;
    rdbk = XmCreateLabel(legendBB,"rdbk",args,n);
    XtVaGetValues(rdbk,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(rdbk,XmNx,rdbkX,XmNy,MARGIN,NULL);
    XtManageChild(rdbk);
    XmStringFree(labelString);

/* average */
    labelString = XmStringCreateSimple(legendString[AVERAGE]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,averageX); n++;
    average = XmCreateLabel(legendBB,"average",args,n);
    XtVaGetValues(average,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(average,XmNx,averageX,XmNy,MARGIN,NULL);
    XtManageChild(average);
    XmStringFree(labelString);

/* rms */
    labelString = XmStringCreateSimple(legendString[RMS]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,rmsX); n++;
    rms = XmCreateLabel(legendBB,"rms",args,n);
    XtVaGetValues(rms,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(rms,XmNx,rmsX,XmNy,MARGIN,NULL);
    XtManageChild(rms);
    XmStringFree(labelString);

/* diff5 */
    labelString = XmStringCreateSimple(legendString[DIFF5]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,diff5X); n++;
    diff5 = XmCreateLabel(legendBB,"diff5",args,n);
    XtVaGetValues(diff5,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(diff5,XmNx,diff5X,XmNy,MARGIN,NULL);
    XtManageChild(diff5);
    XmStringFree(labelString);

/* diff */
    labelString = XmStringCreateSimple(legendString[DIFF]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,diffX); n++;
    diff = XmCreateLabel(legendBB,"diff",args,n);
    XtVaGetValues(diff,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(diff,XmNx,diffX,XmNy,MARGIN,NULL);
    XtManageChild(diff);
    XmStringFree(labelString);

/* units */
    labelString = XmStringCreateSimple(legendString[UNITS]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,unitsX); n++;
    units = XmCreateLabel(legendBB,"units",args,n);
    XtVaGetValues(units,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(units,XmNx,unitsX,XmNy,MARGIN,NULL);
    XtManageChild(units);
    XmStringFree(labelString);

/* operating range */
    labelString = XmStringCreateSimple(legendString[OPER_RANGE]);
    n = 0;
    XtSetArg(args[n],XmNlabelString,labelString); n++;
    XtSetArg(args[n],XmNx,operRangeX); n++;
    operRange = XmCreateLabel(legendBB,"operRange",args,n);
    XtVaGetValues(operRange,XmNwidth,&labelWidth,NULL);
    XtVaSetValues(operRange,XmNx,operRangeX,XmNy,MARGIN,NULL);
    XtManageChild(operRange);
    XmStringFree(labelString);

  }


  XtManageChild(mainBB1);	/* since this could be first create... */

  XtVaGetValues(mainBB1,XmNwidth,&width,NULL);
  XtVaSetValues(mainShell,XmNwidth,width+MW_MARGIN,NULL);

  XFlush(display);


/* add CA connection handling */
  dpStruct = localListHead;

  while (dpStruct != localListTail->next) {

    if (dpStruct->entryType == ProcessVariable) {
/*
 * ProcessVariable
 */
      dpStruct->foundSetPoint = False;
      /* not really treating setPoint as monitorable, but need info */
	  SEVCHK(ca_build_and_connect(dpStruct->channel,TYPENOTCONN,0,
		&(dpStruct->setPointChid),NULL,processSetPointConnectionEvent,
		NULL), "createDisplayPage: error in ca_build_and_connect");
      if (dpStruct->setPointChid != NULL)
		ca_set_puser(dpStruct->setPointChid,dpStruct);
      SEVCHK(ca_build_and_connect(dpStruct->channel,
		TYPENOTCONN,0,
		&(dpStruct->readbackChid),NULL,processReadbackConnectionEvent,
		NULL), "createDisplayPage: error in ca_build_and_connect");
      if (dpStruct->readbackChid != NULL)
		ca_set_puser(dpStruct->readbackChid,dpStruct);
    }

    dpStruct = dpStruct->next;

   }

   ca_pend_event(CA_PEND_EVENT_TIME);

/* and scroll the window in case people are dropping into emptied window */
   if (scrollToWidget != (Widget)NULL)
	XmScrollVisible(mainMW,scrollToWidget,2,2);

   XUndefineCursor(display,XtWindow(mainMW));
}



void renderStatus(DpStruct *dpStruct)
{
  GC gc=0;
  Arg args[2];
  Dimension height;

  Boolean fail = True;

  if (dpStruct != (DpStruct *)NULL)
    if (dpStruct->DA != (Widget)NULL)
      if (XtWindow(dpStruct->DA) != (Window)NULL)
	fail = False;

/* don't care about status if no statusChid or statusEvid */
  if (dpStruct->statusChid == NULL || dpStruct->statusEvid == NULL)
	fail = True;

  if (!fail && dpStruct->status != NoStatus) {

    switch (dpStruct->status) {
      case BadStatus:
        gc = badStatusGC;
        break;
      case GoodStatus:
        gc = goodStatusGC;
        break;
    }

    XtSetArg(args[0],XmNheight,&height);
    XtGetValues(dpStruct->DA,args,1);

    /* inner */
    XFillArc(display,XtWindow(dpStruct->DA),gc,
	globalStatusX,(int)(height-STATUS_DIAM)/2,
	STATUS_DIAM,STATUS_DIAM,0,360*64);

    /* border */
    XDrawArc(display,XtWindow(dpStruct->DA),borderGC,
	globalStatusX,(int)(height-STATUS_DIAM)/2,STATUS_DIAM,STATUS_DIAM,0,360*64);

  }
}

