/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include "dp.h"

#include <Xm/AtomMgr.h>
#include <Xm/DragDrop.h>
#include <ctype.h>

extern void insertElementAfter(char *, DpStruct *);
extern void moveElementsAfter(DpStruct *, DpStruct *, Boolean);
extern DpStruct *draggedDpStruct;
extern void removeElements(DpStruct *, Boolean);

void handleTransfer(Widget w, XtPointer closure, Atom *seltype,
  Atom *type, XtPointer value,  unsigned long *length, int *format);
void handleTrashTransfer(Widget w, XtPointer closure, Atom *seltype,
  Atom *type, XtPointer value, unsigned long *length, int *format);
void handleDrop(Widget w, XtPointer client_data, XtPointer call_data);
void addDropSite(Widget w);
void handleTrashDrop(Widget w, XtPointer client_data, XtPointer call_data);
void addTrashDropSite(Widget w);
void getCompositeHeadTail(DpStruct *move, DpStruct **head, DpStruct **tail);
void moveElementsAfter(DpStruct *move, DpStruct *after, Boolean recurse);
Boolean isCurrentlySelected(DpStruct *dpStruct);
void removeElements(DpStruct *move, Boolean recurse);


static Position dropX, dropY;


/*
 * handle transfer callback
 */
void handleTransfer(Widget w, XtPointer closure, Atom *seltype,
  Atom *type, XtPointer value,  unsigned long *length, int *format)
{
  Arg		args[10];
  Position	daY;
  Dimension	daHeight;
  DpStruct	*dpStruct, *savedDpStruct;
  char		*localString;


  if (*type == COMPOUND_TEXT) {

   savedDpStruct = listTail;
  /* dropY seems to be in the right coordinates for direct use looking up DAs */
   XtSetArg(args[0],XmNy,&daY);
   XtSetArg(args[1],XmNheight,&daHeight);
   dpStruct = listHead;
   while (dpStruct != NULL) {
     XtGetValues(dpStruct->DA,args,2);
     if (dropY >= daY && dropY <= (Position)(daY+daHeight)) {
	savedDpStruct = dpStruct;
	break;
     }
       dpStruct = dpStruct->next;
   }

   if (draggedDpStruct == (DpStruct *)NULL) {
/*
 * normal "create" type drop
 */
  /* strip leading whitespace */
     localString = (char *)value;
     while (isspace(localString[0])) localString++;
    /* add new entry to display */
     insertElementAfter(localString,savedDpStruct);

   } else {
/*
 * since we have a non-null draggedDpStruct, this drag originated in DP, and
 *  hence is a move type drop
 */
     moveElementsAfter(draggedDpStruct,savedDpStruct,True);
   }


  }

  /* Free the value if there is one, or we would have a memory leak */
  if (value) XtFree(value);

  draggedDpStruct = (DpStruct *)NULL;
}


/*
 * handle trash transfer callback
 */
void handleTrashTransfer(Widget w, XtPointer closure, Atom *seltype,
  Atom *type, XtPointer value, unsigned long *length, int *format)
{

  if (draggedDpStruct != NULL) removeElements(draggedDpStruct,True);

  /* Free the value if there is one, or we would have a memory leak */
  if (value) XtFree(value);

  draggedDpStruct = (DpStruct *)NULL;
}



/*
 * handle drop callback
 */
void handleDrop(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmDropProcCallback     dropData;
  XmDropTransferEntryRec transferEntries[2];
  XmDropTransferEntry    transferList;
  Arg                    args[6];
  int                    n;


  dropData = (XmDropProcCallback) call_data;
  dropX = dropData->x;
  dropY = dropData->y;

  /* set the transfer resources */
  n = 0;

  /* if the action is not Drop or the operation is not Copy || Move,
   * cancel the drop */

  if ((dropData->dropAction != XmDROP) ||
      (dropData->operation != XmDROP_COPY &&
       dropData->operation != XmDROP_MOVE)) {
     XtSetArg(args[n], XmNtransferStatus, XmTRANSFER_FAILURE);
     n++;
     /* start the transfer or cancel */
     XmDropTransferStart(dropData->dragContext, args, n);
     draggedDpStruct = (DpStruct *)NULL;
     return;
  } else {
  /* the drop can continue. Establish the transfer list
   * and start the transfer */
     
    transferEntries[0].target = COMPOUND_TEXT;
    transferEntries[0].client_data = (XtPointer)w;
    transferList = transferEntries;
    XtSetArg(args[n],XmNdropTransfers,transferList); n++;
    XtSetArg(args[n],XmNnumDropTransfers,1); n++;
 /* setup transfer proc to accept the drop transfer data */
    XtSetArg(args[n],XmNtransferProc,handleTransfer); n++;
  }

  /* start the transfer or cancel */
  XmDropTransferStart(dropData->dragContext, args, n);
}




void addDropSite(Widget w)
{
  Atom importTargets[2];
  Arg args[8];
  int n;

/* register widget as drop site for CompoundText (as from MEDM) */
  if (COMPOUND_TEXT == (Atom)NULL) {
    COMPOUND_TEXT = XmInternAtom(XtDisplay(w),"COMPOUND_TEXT",False);
  }
  importTargets[0] = COMPOUND_TEXT;
  importTargets[1] = (Atom)NULL;
  n = 0;
  XtSetArg(args[n],XmNimportTargets,importTargets); n++;
  XtSetArg(args[n],XmNnumImportTargets,1); n++;
  XtSetArg(args[n],XmNdropSiteType,XmDROP_SITE_COMPOSITE); n++;
  XtSetArg(args[n],XmNdropProc,handleDrop); n++;
  XmDropSiteRegister(w,args,n);

}


/*
 * handle trash drop callback
 */
void handleTrashDrop(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmDropProcCallback     dropData;
  XmDropTransferEntryRec transferEntries[2];
  XmDropTransferEntry    transferList;
  Arg                    args[6];
  int                    n;


  dropData = (XmDropProcCallback) call_data;
  dropX = dropData->x;
  dropY = dropData->y;

  /* set the transfer resources */
  n = 0;

  /* if the action is not Drop or the operation is not Copy || Move,
   * cancel the drop */

  if ((dropData->dropAction != XmDROP) ||
      (dropData->operation != XmDROP_COPY &&
       dropData->operation != XmDROP_MOVE)) {
     XtSetArg(args[n], XmNtransferStatus, XmTRANSFER_FAILURE);
     n++;
     /* start the transfer or cancel */
     XmDropTransferStart(dropData->dragContext, args, n);
     draggedDpStruct = (DpStruct *)NULL;
     return;
  } else {
  /* the drop can continue. Establish the transfer list
   * and start the transfer */
     
    transferEntries[0].target = COMPOUND_TEXT;
    transferEntries[0].client_data = (XtPointer)w;
    transferList = transferEntries;
    XtSetArg(args[n],XmNdropTransfers,transferList); n++;
    XtSetArg(args[n],XmNnumDropTransfers,1); n++;
 /* setup transfer proc to accept the drop transfer data */
    XtSetArg(args[n],XmNtransferProc,handleTrashTransfer); n++;
  }

  /* start the transfer or cancel */
  XmDropTransferStart(dropData->dragContext, args, n);
}


void addTrashDropSite(Widget w)
{
  Atom importTargets[2];
  Arg args[8];
  int n;

/* register widget as drop site for CompoundText (as from MEDM) */
  if (COMPOUND_TEXT == (Atom)NULL) {
    COMPOUND_TEXT = XmInternAtom(XtDisplay(w),"COMPOUND_TEXT",False);
  }
  importTargets[0] = COMPOUND_TEXT;
  importTargets[1] = (Atom)NULL;
  n = 0;
  XtSetArg(args[n],XmNimportTargets,importTargets); n++;
  XtSetArg(args[n],XmNnumImportTargets,1); n++;
  XtSetArg(args[n],XmNdropSiteType,XmDROP_SITE_COMPOSITE); n++;
  XtSetArg(args[n],XmNdropProc,handleTrashDrop); n++;
  XmDropSiteRegister(w,args,n);

}


void getCompositeHeadTail(DpStruct *move, DpStruct **head, DpStruct **tail)
{
  DpStruct *moveHead, *moveTail;

/*
 * get elements to move (can be single or composite
 */
  moveHead = move;
  moveTail = move;

  if (moveHead->entryType == CompositeDevice && moveTail->next != NULL) {

/* if selected composite, set head and tail pointers appropriately */
  /* march down and get moveTail */
    if (moveTail->next->memberOfComposite) {
      while (moveTail->next != NULL && moveTail->next->memberOfComposite) {
	moveTail = moveTail->next;
      }
    }

  } else if (moveHead->memberOfComposite) {

/* if selected member of composite, set head and tail pointers appropriately */
  /* march back and get proper moveHead */
    if (moveHead->prev != NULL) {
      moveHead = moveHead->prev;
      while (moveHead->prev != NULL && moveHead->entryType != CompositeDevice) {
	moveHead = moveHead->prev;
      }
    }
  /* march down and get moveTail */
    if (moveTail->next != NULL) {
      if (moveTail->next->memberOfComposite) {
        while (moveTail->next != NULL && moveTail->next->memberOfComposite) {
	  moveTail = moveTail->next;
        }
      }
    }

  }
  *head = moveHead;
  *tail = moveTail;
}



/*
 * function to move the list of elements between moveHead and moveTail
 *  (inclusive) to right after afterElement (which should be atomic or
 *  PV or last entry in composite)
 */

void moveElementsAfter(DpStruct *move, DpStruct *after, Boolean recurse)
{
  DpStruct *moveHead, *moveTail, *moveAfter, *dpStruct;
  Dimension rowHeight;
  Arg args[2];
  int j, numSelected;

/*
 * if moving currently selected guys, recursively do all of them
 *   otherwise just do the actual dragged element
 */
  if (recurse) {
    if (isCurrentlySelected(move)) {
      numSelected = getNumberSelected();
      for (j = 0; j < numSelected; j++) {
        moveElementsAfter(currentlySelectedDpStructs[j],after,False);
      }
      return;
    }
  }



/* get head and tail of elements to be moved */
  getCompositeHeadTail(move,&moveHead,&moveTail);

/* now get last element of target to move after (if composite or child) */
  moveAfter = after;
  if ((moveAfter->entryType == CompositeDevice || moveAfter->memberOfComposite)
		&& moveAfter->next != NULL) {
      while (moveAfter->next != NULL && moveAfter->next->memberOfComposite)
	  moveAfter = moveAfter->next;
  }


/* bail out if all this calculation failed */
  if (moveHead == (DpStruct *)NULL || moveHead == moveAfter
				|| moveTail == moveAfter)
	return;


/* move element in list - handle composites (move chains of elements) */

  if (moveHead == listHead) {

    listHead = moveTail->next;
    if (listHead != (DpStruct *)NULL) listHead->prev = (DpStruct *)NULL;

  } else if (moveTail == listTail) {

    listTail = moveHead->prev;
    if (listTail != (DpStruct *)NULL) listTail->next = (DpStruct *)NULL;

  } else {

/* pull moveHead...moveTail out of chain as stand-alone segment then proceed */
    moveHead->prev->next = moveTail->next;
    moveTail->next->prev = moveHead->prev;

  }

  if (moveAfter != NULL)
    if (moveAfter->next != NULL) moveAfter->next->prev = moveTail;
  moveTail->next = moveAfter->next;
  moveHead->prev = moveAfter;
  moveAfter->next = moveHead;

  if (listTail == moveAfter ) listTail = moveTail;
  listTail->next = NULL;


/* now move elements on the screen */
  j = 0;
  dpStruct = listHead;
  if (dpStruct != NULL) {
    XtVaGetValues(dpStruct->setPointTF,XmNheight,&rowHeight,NULL);
    while (dpStruct != (DpStruct *)NULL) {
      XtSetArg(args[0],XmNy,(Position)(j*(rowHeight+MOVE_MARGIN)));
      XtSetValues(dpStruct->DA,args,1);
      j++;
      dpStruct = dpStruct->next;
    }
  }

}


/*
 * function to return T/F if specified widget is in currently selected list
 *   note: for composites, only the composite is in the list, not all
 *   all the members too
 */
Boolean isCurrentlySelected(DpStruct *dpStruct)
{
  DpStruct *ptr;
  int i, numSelected;

  if (currentlySelectedDpStructs != (DpStruct **)NULL) {

/* if selected member of composite,  march back and get proper composite */
    ptr = dpStruct;
    if (dpStruct->memberOfComposite) {
      if (ptr->prev != NULL) {
        ptr = ptr->prev;
        while (ptr->prev != NULL && ptr->entryType != CompositeDevice)
	       ptr = ptr->prev;
      }
    }
    numSelected = getNumberSelected();
    for (i = 0; i < numSelected; i++) {
      if (ptr == currentlySelectedDpStructs[i]) {
	return (True);
      }
    }
  }
  return (False);
}


/*
 * function to remove the list of elements between moveHead and moveTail
 *  (inclusive)
 */
void removeElements(DpStruct *move, Boolean recurse)
{
  DpStruct *moveHead, *moveTail, *dpStruct, *ptr, *savedPtr, *next;
  Dimension rowHeight;
  XClientMessageEvent clientMessageEvent;
  Arg args[2];
  int j, numSelected;

/*
 * if removing currently selected guys, recursively do all of them
 *   otherwise just do the actual dragged element
 */
  if (recurse) {
    if (isCurrentlySelected(move)) {
      numSelected = getNumberSelected();
      for (j = 0; j < numSelected; j++) {
        removeElements(currentlySelectedDpStructs[j],False);
      }
      if (currentlySelectedDpStructs != (DpStruct **)NULL) {
        free(currentlySelectedDpStructs);
        currentlySelectedDpStructs = (DpStruct **)NULL;
      }
      return;
    }
  }



/* get head and tail of elements */
  getCompositeHeadTail(move,&moveHead,&moveTail);

/* bail out if all this calculation failed */
  if (moveHead == (DpStruct *)NULL) return;

/* move element in list - handle composites (move chains of elements) */

  if (moveHead == listHead) {

    listHead = moveTail->next;
    if (listHead != (DpStruct *)NULL) listHead->prev = (DpStruct *)NULL;

  } 
  if (moveTail == listTail) {

    listTail = moveHead->prev;
    if (listTail != (DpStruct *)NULL) listTail->next = (DpStruct *)NULL;

  } 

/* pull moveHead...moveTail out of chain as stand-alone segment then proceed */
  if (moveHead->prev != NULL) (moveHead->prev)->next = moveTail->next;
  if (moveTail->next != NULL) (moveTail->next)->prev = moveHead->prev;


  moveTail->next = 0;
  moveTail->prev = 0;

/* free CA and X resources in the removed elements */
  next = moveTail->next;
  ptr = moveHead;
  while (ptr != next) {
    if (ptr->setPointEvid != NULL) ca_clear_event(ptr->setPointEvid);
    if (ptr->readbackEvid != NULL) ca_clear_event(ptr->readbackEvid);
    if (ptr->setPointChid != NULL) ca_clear_channel(ptr->setPointChid);
    if (ptr->readbackChid != NULL) ca_clear_channel(ptr->readbackChid);

  XtUnmanageChild(ptr->DA);
  XtDestroyWidget(ptr->DA); 


/* 
 * send a clientMessage event to queue up an XtDestroyWidget() call for
 *   processing after the drag drop callbacks have been processed
 */
/* problems with free memory reads in renderStatus (needs dpStruct) when multiple lines cut */
/*
   clientMessageEvent.type = ClientMessage;
   clientMessageEvent.serial = 0;
   clientMessageEvent.send_event = True;
   clientMessageEvent.display = display;
   clientMessageEvent.window = clientMessageWindow;
   clientMessageEvent.message_type = DP_DESTROY;
   clientMessageEvent.format = 32;
   clientMessageEvent.data.l[0] = (long)ptr->DA;

   XSendEvent(display,clientMessageWindow,False,NoEventMask,
	(XEvent *)&clientMessageEvent);
*/

    savedPtr = ptr;
    ptr = ptr->next;
    free(savedPtr);
  }
  ca_pend_event(CA_PEND_EVENT_TIME);

/* now move elements on the screen */
  j = 0;
  dpStruct = listHead;
  if (dpStruct != NULL) {
    XtVaGetValues(dpStruct->DA,XmNheight,&rowHeight,NULL);
    while (dpStruct != (DpStruct *)NULL) {
      XtSetArg(args[0],XmNy,(Position)(j*(rowHeight+MOVE_MARGIN)));
      XtSetValues(dpStruct->DA,args,1);
      j++;
      dpStruct = dpStruct->next;
    }
  }


}
